// MyTreeCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "SQHMTree.h"
#include "MyTreeCtrl.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define MAX_VALUE_LEN 25

#define HOTMETALAPP_GUID  __T( "HoTMetaL PRO.Application" )
#define MAXLEN 512
#define ITEM_REFRESH_TREE 0

#define DOM_ELEMENT	 1
#define DOM_ATTRIBUTE 2
#define DOM_TEXT	3
#define DOM_COMMENT	8
#define DOM_DOCUMENT 9
#define DOM_NOTATION  12

#define IMAGE_ROOT		0
#define IMAGE_DOCUMENT	1
#define IMAGE_COMMENT	2
#define IMAGE_TEXT		3
#define IMAGE_ATTRIBUTE	4
#define IMAGE_ELEMENT	5

/////////////////////////////////////////////////////////////////////////////
// CMyTreeCtrl

CMyTreeCtrl::CMyTreeCtrl()
{
	m_menu.CreatePopupMenu( );
	m_menu.AppendMenu( MF_STRING, ID_REFRESH_TREE, __T("Refresh") );
 
}

BEGIN_MESSAGE_MAP(CMyTreeCtrl, CTreeCtrl)
	//{{AFX_MSG_MAP(CMyTreeCtrl)
	ON_COMMAND( ID_REFRESH_TREE, OnRefreshTree )
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMyTreeCtrl message handlers

////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////
///			Recreate the tree from the document
void CMyTreeCtrl::OnRefreshTree( )
{
	DeleteAllItems( );
	CWaitCursor dummy;
	LoadXMLDocElements( );
}
////////////////////////////////////////////////////////////////////////////////////////////
///			Recreate the tree from the document
void CMyTreeCtrl::OnQuitXMApp( )
{
 	_Application hmetalApp;
	COleException e;
	if (!hmetalApp.CreateDispatch(HOTMETALAPP_GUID, &e)) {
		::AfxMessageBox(IDS_ERR_DISPATCH);
		return;
	}

	hmetalApp.Quit( 0 );


}
////////////////////////////////////////////////////////////////////////////////////////////
///				Set no document open  stub
void CMyTreeCtrl::SetNoDocument( )
{
	TV_INSERTSTRUCT	 tvstruct;
	tvstruct.hParent =  NULL;
	tvstruct.hInsertAfter = TVI_SORT;
	tvstruct.item.iImage = 0;
	tvstruct.item.iSelectedImage = 0;
	tvstruct.item.pszText = __T("(no document open)"); 
	tvstruct.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_TEXT | TVIF_PARAM;
	InsertItem( &tvstruct );
}

////////////////////////////////////////////////////////////////////////////////////////////
//			Load XML document's element to the tree
BOOL CMyTreeCtrl::LoadXMLDocElements( )
{
	// Connect to XMetaL
	COleException e;
	_Application hmetalApp;
	if (!hmetalApp.CreateDispatch(HOTMETALAPP_GUID, &e)) {
		::AfxMessageBox(IDS_ERR_DISPATCH);
		return S_FALSE;
	}
	_Document hmetalDoc = hmetalApp.GetActiveDocument( );

	// If not document open, display stub
	if (!hmetalDoc ) {
        
		// Insert a empty root item when no active document is open
		SetNoDocument( );

		// tvstruct.item.lParam = 0;  // TODO This is useless code?

		// Disable "Refresh" context menu
		// TODO Get rid of this.  Else we get into a state where we can't refresh?
 		// m_menu.EnableMenuItem( ITEM_REFRESH_TREE, FALSE );

        return S_FALSE;
    }

	DOMNode domEle = hmetalDoc.GetDocumentElement( );

	// If no document element available, display stub
	if (!domEle ) {
        
		// Insert a empty root item when no active document is open
		TV_INSERTSTRUCT	 tvstruct;
		tvstruct.hParent =  NULL;
		tvstruct.hInsertAfter = TVI_SORT;
		tvstruct.item.iImage = 0;
		tvstruct.item.iSelectedImage = 0;
		tvstruct.item.pszText = __T("(empty)"); 
		tvstruct.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_TEXT | TVIF_PARAM;
		InsertItem( &tvstruct );
		// tvstruct.item.lParam = 0;  // TODO This is useless code?

		// Disable "Refresh" context menu
		// TODO Get rid of this.  Else we get into a state where we can't refresh?
 		// m_menu.EnableMenuItem( ITEM_REFRESH_TREE, FALSE );

        return S_FALSE;
    }

	// Insert the element names into the tree, starting with the root
	int index = 0;
	InsertElement( NULL, domEle, index );

	// Enable "Refresh" context menu
	m_menu.EnableMenuItem( ITEM_REFRESH_TREE, TRUE );

	// Expand the tree to the first level
	Expand(GetRootItem(), TVE_EXPAND);

	return TRUE;
}

/*///////////////////////////////////////////////////////////////////////////////////////////
  NAME
        InsertElement - insert an element into the tree

	ARGUMENTS
		root - root node of tree control
		element - DOM node to insert, if NULL, does nothing
		index - ??
///////////////////////////////////////////////////////////////////////////////////////////*/

void  CMyTreeCtrl::InsertElement( HTREEITEM root, DOMNode element, int & index )
{
	// If DOM element is NULL, do nothing
	// Handles case where document is empty, or plain text view
	if( element == NULL ) return;

	TV_INSERTSTRUCT	 tvstruct;
 
	tvstruct.hParent =  root;
	tvstruct.hInsertAfter = TVI_LAST;

	if( root == NULL ){		// root item
		tvstruct.item.iImage =  0;
		tvstruct.item.iSelectedImage = 0;
	}
	else{	// non-root item
		tvstruct.item.iImage = ItemImageID( element.GetNodeType( ) );
		tvstruct.item.iSelectedImage = ItemImageID( element.GetNodeType( ) );
	}

	CString name =  element.GetNodeName( );
	short type = element.GetNodeType();
	if (type != 1 /* element */) {
		//++index;
		return;
	}
	CString value;
	if (type == 1 /* element */) {
		value = GetTextContent(element);
		value = NormalizeWhiteSpace(value);
	}
	else {
		value = "";
	}

	if( !value.IsEmpty( ) ){
		if( value.GetLength( ) > MAX_VALUE_LEN ){
			value = value.Left( MAX_VALUE_LEN - 3 ) + "...";
		}
	}
	name = __T( "<" ) + name + __T("> ") + value; 

	char item[ MAXLEN ];

	for( int  i = 0; i < name.GetLength( ); ++i )
		item[ i ] = (char) name[ i ];

	item[i] = '\0';


	tvstruct.item.pszText =  item; 
	tvstruct.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_TEXT | TVIF_PARAM;
	tvstruct.item.lParam = index;
	++index;

	HTREEITEM root1;

    root1 = InsertItem( &tvstruct );

	DOMNodeList list = element.GetChildNodes( );

	// iterate through all the child nodes
	for( i = 0; i < list.GetLength( ); ++i )
		InsertElement( root1,  list.item( i ), index );
}

///////////////////////////////////////////////////////////////////////////////////////////
///			return the image id to be set for a specific item type
int CMyTreeCtrl::ItemImageID( int type )
{
	switch( type ){
		case DOM_ELEMENT:	return IMAGE_ELEMENT;
		case DOM_ATTRIBUTE:	return IMAGE_ATTRIBUTE;
		case DOM_COMMENT:	return IMAGE_COMMENT;
		case DOM_TEXT:	return IMAGE_TEXT;
		case DOM_DOCUMENT:	return IMAGE_DOCUMENT;
		default:		return IMAGE_ELEMENT;
	}
}

////////////////////////////////////////////////////////////////////
//  NAME
//      GetTextContent - return text content of an element as string
//
//  HISTORY
//      1999.02.27  Created by Bruce Sharpe
//
//  Copyright (c) 1999 SoftQuad Software Inc.  All rights reserved.
////////////////////////////////////////////////////////////////////

CString CMyTreeCtrl::GetTextContent(DOMNode elt)
{
    CString str = "";
    DOMNodeList nodeList = elt.GetChildNodes();
    long numChildren = nodeList.GetLength();
    for (long child = 0; child < numChildren; child++) {
        DOMNode node = nodeList.item(child);
        if (node.GetNodeType() != 3 /* text */) break;
        str += node.GetNodeValue();
    }
    return str;
}

CString CMyTreeCtrl::NormalizeWhiteSpace(CString str)
{
    LPTSTR pIn, pOut;
    bool bLastWasSpace;

    // Get the character buffer associated with the CString
    LPTSTR pBuf = str.GetBuffer(256);
    size_t len = strlen(pBuf);

    // First pass: convert all white space to ordinary space
    pIn = pBuf;
    for (size_t index = 0; index < len; index++) {
        if (isspace(*pIn)) *pIn = ' ';
        pIn++;
    }

    // Second pass: collapse all repeated space to single space
    // pOut points to the next place to copy a character to
    pIn = pOut = pBuf;
    bLastWasSpace = false;
    for (index = 0; index < len; index++) {
        if (bLastWasSpace) {
            if (' ' == *pIn) pIn++;
            else {
                *pOut++ = *pIn++;
                bLastWasSpace = false;
            }
        }
        else {
            if (' ' == *pIn) {
                *pOut++ = *pIn++;
                bLastWasSpace = true;
            }
            else {
                *pOut++ = *pIn++;
            }
        }
    }
    *pOut = '\0';
    
    // Release the character buffer
    str.ReleaseBuffer(-1);  // -1 because our string is null-terminated

    // Remove any white space from beginning or end of string
    str.TrimLeft();
    str.TrimRight();
    
    return str;
}

//	DESCRIPTION
//	If user has selected a tree item with left mouse button, navigate
//	to that point in the document.
//
//	NOTES
//	The use of getElementsByTagName() in the code is efficient, but it
//	restricts the algorithm to work with elements only.  If the tree also
//	holds other types of nodes, a different algorithm is required.

void CMyTreeCtrl::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CTreeCtrl::OnLButtonDown(nFlags, point);

	UINT flag;
	HTREEITEM  hItem = HitTest( point, &flag );
	HTREEITEM hItemSelected = GetSelectedItem( );

	if(hItem && hItemSelected && (hItem == hItemSelected) && (flag | TVHT_ONITEM) ){

		// Connect to XMetaL
		COleException e;
		_Application hmetalApp;
		if (!hmetalApp.CreateDispatch(HOTMETALAPP_GUID, &e)) {
			::AfxMessageBox(IDS_ERR_DISPATCH);
			return;
		}
		
		_Document hmetalDoc = hmetalApp.GetActiveDocument( );


		//Get the struct of selected item
		TV_ITEM item;
		item.hItem = hItem;
		item.mask = TVIF_PARAM;
		int nRet = GetItem( &item );

		int index = (int) item.lParam;

		if( !hmetalDoc  ){
			::AfxMessageBox( IDS_ERR_NOACTIVEDOC  );
			return;
		}
		else{

			Range rng = hmetalDoc.GetRange();
			Selection selection = hmetalApp.GetSelection( );

			DOMNodeList nodeList = hmetalDoc.getElementsByTagName("*");
			if (nodeList.GetLength() > index) {
				DOMNode node = nodeList.item(index);
				rng.SelectNodeContents( node );
				rng.Collapse(1);
				rng.Select();
				hmetalDoc.ScrollToSelection( );
			}
		}
	}
}

// DESCRIPTION
//	Display context menu

void CMyTreeCtrl::OnRButtonDown(UINT nFlags, CPoint point) 
{
	ClientToScreen(&point);
	m_menu.TrackPopupMenu(TPM_CENTERALIGN | TPM_LEFTBUTTON, 
		point.x, point.y, this);
	CTreeCtrl::OnRButtonDown(nFlags, point);
}
